class UserLoginException(Exception):
    """Exception when logging in"""

    pass


class MealFetchingException(Exception):
    """Exception when no meal ids fetched"""

    pass


class ChangingMealsException(Exception):
    """Exception when unable to change meal"""

    pass

class InsufficientVariables(Exception):
    """Exception when user has not provided enaugh info"""

    pass



"""
eAsistentUser
=============

eAsistentUser class is used for creating and accessing uer's data.
For examples and in depth usage, take a look at our documentation.
<https://dasadweb.com/documentation/eAwaiter>
"""

from dataclasses import dataclass
from datetime import datetime

@dataclass
class eAsistentUser():
    """
    username, password => user's eAsistent credentials
    disliked_foods, favourite_foods => 
        array of foods (str spelled exactly as in eAsistent)
    preferred_menu, default_menu =>
        integers of menu options
        0 = odjava (sign off)
        1-6 = meal options

        1 ... Topli Mesni
        2 ... Topli Brezmesni
        3 ... Hladni Mesni
        4 ... Hladni Brezmesni
        5 ... Solatni
        6 ... Šport
    overwrite_unchangable =>
        user is unable to change meals < 7 days in advance
        when True:
            this option will try to change even if < 7 days in advance
    """
    username: str
    password: str

    disliked_foods: list
    favourite_foods: list

    preferred_menu: int
    default_menu: int

    overwrite_unchangable: bool

@dataclass
class eAsistentMeal():
    """
    meal_text => contents of meal
    meal_id => id of meal (not same as eAsistentUser menu)
    date => date of menu
    changable => true if > 8 days in advance
    selected => true if meal selected
    """
    meal_text: str
    meal_id: str
    date: datetime
    changable: bool
    selected: bool

"""
Created By:

|~| ._ _'|
_)|<| (_ |
"""




"""
Helpers
=======

Helpers are different functions that just don't belong in any of the classes.
For examples and in depth usage, take a look at our documentation.
<https://dasadweb.com/documentation/eAwaiter>
"""

from datetime import datetime, timedelta


def weeks_in_advance(weeks: int, first_week_school: datetime) -> [int, datetime]:
    today = datetime.now()
    # if weekends => getmonday | fetches another week in advance = error
    # if today.strftime("%w") == "0":
    #     today -= timedelta(days=6)
    # elif today.strftime("%w") == "6":
    #     today -= timedelta(days=5)
    monday = get_monday(today + timedelta(days=weeks * 7))  # gets weeks in advance
    return int(str((monday - first_week_school) / 7).split(" ")[0]), monday


def get_monday(date: datetime = datetime.now()) -> datetime:
    first_day_num = int(date.strftime("%w"))
    # if not 1 (monday) => need to find monday
    if first_day_num == 1:  # 1 == monday
        return date

    # if weekend => goes forwards
    if first_day_num == 6:  # 6 == sat
        monday = date + timedelta(days=2)  # goes to mon
    elif first_day_num == 0:  # 0 == sun
        monday = date + timedelta(days=1)  # goes to mon
    # if workday => goes backwards
    else:  # if [tue-fri]
        monday = date - timedelta(days=first_day_num - 1)  # goes to mon
    return monday


def get_selected(week_meal_data):
    selected_meals: list = []
    for i in range(5):  # each day
        for meal in week_meal_data[i]:  # each meal
            if meal.selected:
                selected_meals.append(meal)
    return selected_meals

"""
Created By:

|~| ._ _'|
_)|<| (_ |
"""




"""
API Navigation
==============

API Navigation is made to purely interact with the api and scrape the meal off the web.
For examples and in depth usage, take a look at our documentation.
<https://dasadweb.com/documentation/eAwaiter>
"""

from bs4 import BeautifulSoup
from bs4.element import Tag

from datetime import datetime, timedelta
import requests

from .dataclasses import eAsistentMeal



HTML_ID = "ednevnik-seznam_ur_teden"


class ApiNavigator:
    session = requests.Session()
    def login(self, username: str, password: str) -> requests.models.Response:
        login_url = "https://www.easistent.com/p/ajax_prijava"
        data = {
            "uporabnik": username,
            "geslo": password,
            "pin": "",
            "captcha": "",
            "koda": "",
        }
        login = self.session.post(login_url, data=data)
        return login

    def get_menu_ids(self) -> list:
        meals_url = "https://www.easistent.com/dijaki/ajax_prehrana_obroki_seznam"
        response = self.session.get(meals_url)
        soup = BeautifulSoup(response.content, "html.parser")
        meal_types = soup.find(class_=HTML_ID).find_all("tr")
        meal_types.pop(0)
        meal_types.pop(-1)

        meal_ids = []
        for meal_type in meal_types:
            meal_id_full = meal_type.find_all("td")[1].get("id")
            meal_id = meal_id_full.split("-")[4]
            meal_ids.append(meal_id)

        return meal_ids

    def get_meal_data(self, week_num: int, monday: datetime, meals: list) -> list:
        data = {
            "qversion": 1,
            "teden": week_num,
            "smer": "naprej",
        }
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        meals_url = "https://www.easistent.com/dijaki/ajax_prehrana_obroki_seznam"
        site = self.session.post(meals_url, data=data, headers=headers)
        soup = BeautifulSoup(site.content, "html.parser")

        work_days = 5
        week_data = []
        for i in range(work_days):
            date = monday + timedelta(days=i)

            day_data = self.get_daily_meals(soup, meals, date)
            week_data.append(day_data)

        return week_data

    def get_daily_meals(self, soup: BeautifulSoup, meals: list, date: datetime) -> list:
        date_str = date.strftime("%Y-%m-%d")
        day_data = []
        for meal_id in meals:
            meal_html_id = f"{HTML_ID}-td-malica-{meal_id}-{date_str}-0"  # html class
            meal_container = soup.find("td", id=meal_html_id).find("div")

            check_first = self.check_single_div(meal_container, meal_id, date_str)
            if check_first[0]:
                day_data.append(
                        check_first[1]
                        )
                continue

            check_second = self.check_double_div(meal_container, meal_id, date, date_str)
            if check_second[0]:
                day_data.append(
                        check_second[1]
                        )
                continue

            check_third = self.check_tripple_div(meal_container, meal_id, date, date_str)
            if check_third[0]:
                day_data.append(
                        check_third[1]
                        )
                continue

        return day_data

    def check_single_div(self, container: Tag, meal_id: str, date_str: str) -> [bool, dict]:
        """
        Only in past meals without meal_text
        """
        if not len(container.find_all("div")) == 1:
            return False, {}
        return True, eAsistentMeal("", meal_id, date_str, False, False)

    def check_double_div(self, container: Tag, meal_id: str, date: datetime, date_str: str) -> [bool, dict]:
        """
        Either: past (not selected),
                past (selected),
                future (not loaded)
        """
        if not len(container.find_all("div")) == 2:
            return False, {}

        selected = False
        changable = False
        second_div = container.find_all("div")[1]

        if len(list(container.findChildren(recursive=False))) > 2:
            # selected or signed off
            selected = True \
                    if container.find("span").text == "Prijavljen" \
                    else False

        meal_text = second_div.text.strip()
        if meal_text.split(" ")[0] in ["Prijava", "Naročen"]:
            # future
            today = datetime.today()
            changable = True \
                    if (date - today).days > 9 \
                    else False
            selected = True \
                    if meal_text.split(" ")[0] == "Naročen" \
                    else False
            meal_text = ""

        return True, eAsistentMeal(meal_text, meal_id, date_str, changable, selected)

    def check_tripple_div(self, container: Tag, meal_id: str, date: datetime, date_str: str) -> [bool, dict]:
        """
        Either: future (not selected),
                future (selected),
                present (selected),
                present (unable to change)
        """
        second_div = container.find_all("div")[1]
        third_div = container.find_all("div")[2]
        third_div_children = third_div.findChildren(recursive=False)
        meal_text = second_div.text.strip()
        selected = False

        if len(list(third_div_children)) > 2:
            selected = True \
                    if third_div.find("span").text == "Naročen" \
                    else False

        today = datetime.today()
        changable = False \
                if third_div_children[0].text == "Izbira menija ni več mogoča" \
                or (date - today).days < 9 \
                else True

        return True, eAsistentMeal(meal_text, meal_id, date_str, changable, selected)

    def prijava_odjava(self, action: str, meal_id: str, date: str) -> requests.models.Response:
        url = "https://www.easistent.com/dijaki/ajax_prehrana_obroki_prijava"
        data = {
            "tip_prehrane": "malica",
            "id_lokacija": "0",
            "akcija": f"{action}",
            "id_meni": f"{meal_id}",
            "datum": f"{date}",
        }
        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
            "X-Requested-With": "XMLHttpRequest",
        }
        response = self.session.post(url, data=data, headers=headers)
        return response


    def send_mail(self, message: str):
        print(message)

"""
Created By:

|~| ._ _'|
_)|<| (_ |
"""




"""
Waiter
======

Waiter class is used for easy interaction with the user.
For examples and in depth usage, take a look at our documentation.
<https://dasadweb.com/documentation/eAwaiter>
"""

from bs4 import BeautifulSoup

from datetime import datetime, timedelta
import json
import os

from .api_navigator import ApiNavigator
from .dataclasses import eAsistentUser
from .exceptions import UserLoginException, MealFetchingException, ChangingMealsException
from .helpers import *


FORMAT = "%Y-%m-%d %H:%M:%S"


class Waiter:
    def __init__(
            self,
            username: str="",
            password: str="",
            disliked_foods: list=[],
            favourite_foods: list=[],
            default_menu: int=1,
            preferred_menu: int=1,
            overwrite_unchangable: bool=False
            ) -> None:

        self.failed_login: bool = False
        self.api: ApiNavigator = ApiNavigator()
        self.user: eAsistentUser = eAsistentUser(
                username=username,
                password=password,
                disliked_foods=disliked_foods,
                favourite_foods=favourite_foods,
                default_menu=default_menu,
                preferred_menu=preferred_menu,
                overwrite_unchangable=overwrite_unchangable
                )

        self.login()

    def login(self) -> None:
        if not self.user.username and not self.user.password:
            raise UserLoginException("Please fill in the Login form!")

        login = self.api.login(
                self.user.username,
                self.user.password
                )

        if login.json()["status"] in ["", "hide_pin"]:
            self.failed_login = True
            # captcha => make them login on web
            if list(login.json()["errfields"].keys())[0] == "captcha":
                raise UserLoginException(
                    "Login failed too many times, please try logging in on the web!"
                )
            raise UserLoginException("Check your credentials and try again!")

        (
            self.meals,
            self.first_week_school
        ) = self.get_school_info()

    def get_school_info(self) -> [list, datetime]:
        date_today = datetime.now()
        curr_month = int(date_today.strftime("%m"))
        curr_year = int(date_today.strftime("%Y"))
        meal_ids = self.api.get_menu_ids()

        school_start_year = curr_year \
                if curr_month >= 9 \
                else curr_year - 1

        first_day_school = datetime(school_start_year, 9, 1)
        first_week_school = get_monday(first_day_school)

        return meal_ids, first_week_school
    
    def prijava_odjava(self, action: str, meal_id: str, date: str):
        result = self.api.prijava_odjava(
            action,
            meal_id,
            date
            )
        print(result.json())
        if not result.json()["status"]:
            # automatically won't work if 7 days from today
            result = self.api.prijava_odjava(
                "odjava",
                meal_id,
                date
                )
        return True if result.json()["status"] else False

    def change_disliked(self, week: int, count: int=1) -> list:
        changes = []
        next_week_num, monday = weeks_in_advance(week, self.first_week_school)
        meals_data = get_selected(
            self.api.get_meal_data(next_week_num, monday, self.meals)
        )

        check_odjava = False
        if int(self.user.preferred_menu) == 0:
            check_odjava = True
        selected_menu = self.meals[int(self.user.preferred_menu) - 1]

        for meal_data in meals_data:
            for disliked_food in self.user.disliked_foods:
                if not meal_data.changable and not self.user.overwrite_unchangable:
                    break
                if disliked_food.upper() in meal_data.meal_text:
                    if meal_data.meal_id == selected_menu:
                        check_odjava = True
                        break
                    if check_odjava: 
                        self.api.prijava_odjava(
                            "odjava", meal_data.meal_id, meal_data.date
                        )
                        changes.append(f"Meal changed for {meal_data.date} (sigend off)!")
                        break

                    response = self.api.prijava_odjava(
                        "prijava", selected_menu, meal_data.date
                    )
                    change = f"Meal changed for {meal_data.date} (to {selected_menu})!"
                    if not response.json()["status"]:
                        response = self.api.prijava_odjava(
                            "odjava", meal_data.meal_id, meal_data.date
                        )
                        change = f"Meal signed off for {meal_data.date}!"
                    changes.append(change)
                    break
        if count < 2:
            count += 1
            self.change_disliked(week, count)
        return changes

    def change_favourites(self, week: int) -> list:
        changes = []
        next_week_num, monday = weeks_in_advance(week, self.first_week_school)
        meals_data = self.api.get_meal_data(next_week_num, monday, self.meals)

        check_odjava = False
        if int(self.user.default_menu) == 0:
            check_odjava = True
        selected_menu = self.meals[int(self.user.default_menu) - 1]

        for i in range(5):
            is_subbed = False
            for meal_data in meals_data[i]:
                for favourite_food in self.user.favourite_foods:
                    if (
                        not meal_data.changable
                        and not self.user.overwrite_unchangable
                        ):
                        break
                    if favourite_food.upper() in meal_data.meal_text:
                        response = self.api.prijava_odjava(
                            "prijava", meal_data.meal_id, meal_data.date
                        )
                        if response.json()["status"]:
                            is_subbed = True
                            changes.append(
                                f"Favourite meal found ({meal_data.date})!"
                                )
                        break
                if is_subbed:
                    break
            if is_subbed:
                continue
            if check_odjava:
                response = self.api.prijava_odjava(
                    "odjava", meals_data[i][0].meal_id, meals_data[i][0].date
                )
                continue
            response = self.api.prijava_odjava(
                "prijava", selected_menu, meals_data[i][0].date
            )
            if not response.json()["status"]:
                response = self.api.prijava_odjava(
                    "odjava", meals_data[i][0].meal_id, meals_data[i][0].date
                )
        return changes

    def get_meals(self, week_num: int = 0, date: datetime = datetime.today()) -> list:
        if week_num:
            print(week_num)
            date = get_monday(self.first_week_school + timedelta(days=7*week_num))
            return self.api.get_meal_data(week_num, date, self.meals)
        monday = get_monday(date)
        week_num = int(str((monday - self.first_week_school) / 7).split(" ")[0])
        return self.api.get_meal_data(week_num, monday, self.meals)

    def service(self) -> None:
        self.api.send_mail("Service started!")
        changes = []

        if self.failed_login:
            raise UserLoginException("Failed to login!")

        try:
            if not self.meals:
                raise MealFetchingException("No meals' ids!")

            if self.user.favourite_foods:
                changes += self.change_favourites(1)

            if self.user.disliked_foods and self.user.preferred_menu:
                changes += self.change_disliked(1)
                
            if self.user.favourite_foods and self.user.default_menu:
                changes += self.change_favourites(2)

            if self.user.disliked_foods:
                changes += self.change_disliked(2)

            self.api.send_mail("service did well")

        except Exception as e:
            self.api.send_mail(e)
        
        return changes

"""
Created By:

|~| ._ _'|
_)|<| (_ |
"""

